/*
 *  Copyright (C) 2006-2010 Leandro Nini
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "mainWindow.h"
#include "AboutDialog.h"
#include "infoDialog.h"
#include "utils.h"
#include "settings.h"
#include "iconFactory.h"
#include "inputFactory.h"
#include "ioEx.h"
#include "trackList/trackListFactory.h"

#include "ledRed.xpm"
#include "ledYellow.xpm"
#include "ledGreen.xpm"

#ifdef WIN32
#  include "singleApp.h"
#endif

FXIMPLEMENT(timeDisplay, FX7Segment, 0, 0)
FXIMPLEMENT(timeLabel, FXLabel, 0, 0)

/*****************************************************************/

FXint loadThread::run() {

	FXint res=_i->open(_name);
	_loaded->setData((void*)(FXival)res);
	_loaded->signal();
	return 0;
}

/*****************************************************************/

FXDEFMAP(mainWindow) mainWindowMap[] = {
	// Message_Type			ID				Message_Handler
	FXMAPFUNC(SEL_IO_READ,		mainWindow::ID_AUDIO_MSG,	mainWindow::onAudioMsg),
	FXMAPFUNC(SEL_UPDATE,		mainWindow::ID_FILETYPE,	mainWindow::onUpdFiletype),
	FXMAPFUNC(SEL_UPDATE,		mainWindow::ID_LED,		mainWindow::onUpdLed),
	FXMAPFUNC(SEL_UPDATE,		mainWindow::ID_PLSONG,		mainWindow::onUpdSubtune),
	FXMAPFUNC(SEL_UPDATE,		mainWindow::ID_SUBTUNE,		mainWindow::onUpdSubtune),
	FXMAPFUNC(SEL_CHANGED,		mainWindow::ID_DIRLIST,		mainWindow::onCmdDirList),
	FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,mainWindow::ID_DIRLIST,	mainWindow::onRgtClkDirList),
	FXMAPFUNC(SEL_DOUBLECLICKED,	mainWindow::ID_PLAYLIST,	mainWindow::onCmdPlayPauseSong),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_PLAYPAUSE,	mainWindow::onCmdPlayPauseSong),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_STOP,		mainWindow::onCmdStopSong),
	FXMAPFUNCS(SEL_COMMAND,		mainWindow::ID_PREV,mainWindow::ID_NEXT,	mainWindow::onCmdChangeSong),
	FXMAPFUNC(SEL_SELECTED,		mainWindow::ID_PLAYLIST,	mainWindow::onCmdSongSelected),
	FXMAPFUNC(SEL_IO_READ,		mainWindow::ID_SONG_LOADED,	mainWindow::onCmdSongLoaded),
	FXMAPFUNCS(SEL_COMMAND,	mainWindow::ID_PREV_SUBTUNE,mainWindow::ID_NEXT_SUBTUNE,	mainWindow::onCmdSubtune),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_BOOKMARKS,	mainWindow::onCmdBmGet),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_OPT_CONFIG,	mainWindow::onCmdConfig),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_FILETYPE,	mainWindow::onCmdFiletype),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_VOL,		mainWindow::onCmdVol),
	FXMAPFUNC(SEL_CHANGED,		mainWindow::ID_VOL,		mainWindow::onCmdVol),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_UI_MODE,		mainWindow::onCmdUIMode),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_PL_BACK,		mainWindow::onCmdPlBack),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_PL_EDIT,		mainWindow::onCmdPlEdit),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_PL_SAVE,		mainWindow::onCmdPlSave),
	FXMAPFUNCS(SEL_COMMAND,		mainWindow::ID_BM_ADD,mainWindow::ID_PL_ADD,	mainWindow::onCmdAddItem),
	FXMAPFUNCS(SEL_COMMAND,		mainWindow::ID_INFO,mainWindow::ID_ABOUT,	mainWindow::onCmdInfo),
	FXMAPFUNC(SEL_COMMAND,		mainWindow::ID_HOME,		mainWindow::onHome),
	FXMAPFUNC(SEL_IO_READ,		mainWindow::ID_MESSAGE,		mainWindow::onMessage),
};

FXIMPLEMENT(mainWindow, FXMainWindow, mainWindowMap, ARRAYNUMBER(mainWindowMap))

mainWindow::mainWindow(FXApp* a) :
	FXMainWindow(a, PACKAGE_NAME),
	_input(0),
	_preload(0),
	_loadThread(0),
	_playing(false),
	_preloaded(FXString::null),
	_playDir(FXString::null) {

	ledRedIcon=new FXXPMIcon(a, ledRed);
	ledYellowIcon=new FXXPMIcon(a, ledYellow);
	ledGreenIcon=new FXXPMIcon(a, ledGreen);

	setIcon(GET_ICON(icon_logo32));
	setMiniIcon(GET_ICON(icon_logo16));

	SETTINGS->load();

	_audio=new audio(this, mainWindow::ID_AUDIO_MSG);

	_toolbar=new FXToolBar(this, LAYOUT_FILL_X);
	FXToolBar* toolbar2=new FXToolBar(this, LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FILL_X);
	_statusbar=new FXStatusBar(this, LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER);
	_statusbar->getStatusLine()->setNormalText(FXString(PACKAGE_STRING));

	// toolbar buttons
	new FXButton(_toolbar, tr("\tPrevious\tGo to previous song"), GET_ICON(icon_mediaskipbackward),
		this, mainWindow::ID_PREV, BUTTON_TOOLBAR|FRAME_RAISED);
	new FXButton(_toolbar, tr("\tStop\tStop playing"), GET_ICON(icon_mediaplaybackstop),
		this, mainWindow::ID_STOP, BUTTON_TOOLBAR|FRAME_RAISED);
	_playpause=new FXButton(_toolbar, tr("\tPlay\tPlay selected song"), GET_ICON(icon_mediaplaybackstart),
		this, mainWindow::ID_PLAYPAUSE, BUTTON_TOOLBAR|FRAME_RAISED);
	new FXButton(_toolbar, tr("\tNext\tGo to next song"), GET_ICON(icon_mediaskipforward),
		this, mainWindow::ID_NEXT, BUTTON_TOOLBAR|FRAME_RAISED);
	_timeDisplay=new timeDisplay(_toolbar, this, ID_TIME);
	new FXToggleButton(_toolbar, tr("\tCompact mode\tSwitch to compact UI mode"),
		tr("\tFull mode\tSwitch to full UI mode"), GET_ICON(icon_viewcompact), GET_ICON(icon_viewfull),
		this, mainWindow::ID_UI_MODE, TOGGLEBUTTON_NORMAL|TOGGLEBUTTON_TOOLBAR);
	new FXButton(_toolbar, tr("\tSettings\tProgram settings"), GET_ICON(icon_preferencesdesktop),
		this, mainWindow::ID_OPT_CONFIG, BUTTON_TOOLBAR|FRAME_RAISED);

	// toolbar
	new FXButton(toolbar2, tr("\tPrevious subtune\tGo to previous subtune"), GET_ICON(icon_goprevious),
		this, mainWindow::ID_PREV_SUBTUNE, BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_CENTER_Y);
	FXLabel* tmpLabel=new FXLabel(toolbar2, tr("\t\tSubtune"), 0, FRAME_SUNKEN|LAYOUT_CENTER_Y);
	tmpLabel->setTarget(this);
	tmpLabel->setSelector(ID_SUBTUNE);
	new FXButton(toolbar2, tr("\tNext subtune\tGo to next subtune"),  GET_ICON(icon_gonext),
		this, mainWindow::ID_NEXT_SUBTUNE, BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_CENTER_Y);

	new FXVerticalSeparator(toolbar2, SEPARATOR_NONE);

	tmpLabel=new FXLabel(toolbar2, FXString::null, ledRedIcon, LABEL_NORMAL|LAYOUT_CENTER_Y);
	tmpLabel->setTarget(this);
	tmpLabel->setSelector(ID_LED);
	tmpLabel->setHelpText(tr("Stop"));

	tmpLabel=new FXLabel(toolbar2, tr("\t\tSong"), 0, FRAME_SUNKEN|LAYOUT_CENTER_Y);
	tmpLabel->setTarget(this);
	tmpLabel->setSelector(ID_PLSONG);

	_playMode=new FXToggleButton(toolbar2, tr("\tPlaylist mode\tSwitch to playlist mode"),
		tr("\tSong mode\tSwitch to song mode"), GET_ICON(icon_song), GET_ICON(icon_playlist),
		0, 0, TOGGLEBUTTON_NORMAL|TOGGLEBUTTON_TOOLBAR|LAYOUT_CENTER_Y);
	FXKnob* volume=new FXKnob(toolbar2, this, mainWindow::ID_VOL, KNOB_NORMAL|LAYOUT_RIGHT);
	volume->setHelpText(tr("Volume"));
	volume->setValue(_audio->volume());

	// statusbar buttons
	new FXButton(_statusbar, tr("\tAbout\tAbout"), GET_ICON(icon_logo16),
		this, mainWindow::ID_ABOUT, BUTTON_TOOLBAR|FRAME_RAISED);
	new FXButton(_statusbar, tr("\tExit\tExit"), GET_ICON(icon_applicationexit),
		this, ID_CLOSE, BUTTON_TOOLBAR|FRAME_RAISED);

	//
	FXHorizontalFrame* hframe=new FXHorizontalFrame(this, LAYOUT_FILL_X|LAYOUT_SIDE_BOTTOM);
	new FXButton(hframe, tr("\tInfo\tSong info"), GET_ICON(icon_helpabout),
		this, mainWindow::ID_INFO, BUTTON_TOOLBAR|FRAME_RAISED);
	_songInfo=new infoLabel(hframe);
	_songTime=new timeLabel(hframe);

	_mainView=new FXHorizontalFrame(this, LAYOUT_FILL|PACK_UNIFORM_WIDTH, 0,0,0,0, 0,0,0,0, 0,0);

	// Left view - file list/bookmarks
	FXVerticalFrame* frame=new FXVerticalFrame(_mainView, FRAME_NORMAL|LAYOUT_FILL, 0,0,0,0, 0,0,0,0, 0,0);
	FXHorizontalFrame* buttons=new FXHorizontalFrame(frame, LAYOUT_FILL_X|PACK_UNIFORM_WIDTH, 0,0,0,0, 0,0,0,0, 0,0);
	FXSwitcher* switcher=new FXSwitcher(frame, LAYOUT_FILL|FRAME_NONE, 0,0,0,0, 0,0,0,0);
	hframe=new FXHorizontalFrame(frame, LAYOUT_FILL_X, 0,0,0,0, 0,0,0,0, 0,0);
	_fileTypes=new FXListBox(hframe, this, mainWindow::ID_FILETYPE, LAYOUT_FILL_X|FRAME_NORMAL);
	_fileTypes->setHelpText(tr("Select backend"));
	_editMode=new FXToggleButton(hframe, tr("\t\tEdit playlist"), "", GET_ICON(icon_editlist),
		GET_ICON(icon_editlist), this, mainWindow::ID_PL_EDIT, TOGGLEBUTTON_KEEPSTATE|TOGGLEBUTTON_NORMAL);
	new FXButton(hframe, tr("\t\tSave playlist"), GET_ICON(icon_documentsave),
		this, mainWindow::ID_PL_SAVE, BUTTON_NORMAL);
	new FXButton(hframe, tr("\t\tReturn to current playlist"), GET_ICON(icon_currentplaylist),
		this, mainWindow::ID_PL_BACK, BUTTON_NORMAL);
	new FXButton(hframe, tr("\t\tHome"), GET_ICON(icon_gohome),
		this, mainWindow::ID_HOME, BUTTON_NORMAL);

	new FXButton(buttons, tr("Playlist"), GET_ICON(icon_playlist), switcher,
		FXSwitcher::ID_OPEN_FIRST, BUTTON_NORMAL|LAYOUT_FILL_X);
	new FXButton(buttons, tr("&Bookmarks"), GET_ICON(icon_bookmark), switcher,
		FXSwitcher::ID_OPEN_SECOND, BUTTON_NORMAL|LAYOUT_FILL_X);

	frame=new FXVerticalFrame(switcher, FRAME_NORMAL|LAYOUT_FILL, 0,0,0,0, 0,0,0,0, 0,0);
	_playList=new playlist(frame, this, mainWindow::ID_PLAYLIST);
	_playList->setNumVisible(10);

	_bookmarkList=new bookmark(switcher, this, mainWindow::ID_BOOKMARKS);
	_bookmarkList->dropEnable();

	// Right view - dir list
	frame=new FXVerticalFrame(_mainView, FRAME_NORMAL|LAYOUT_FILL, 0,0,0,0, 0,0,0,0, 0,0);
	/*buttons=new FXHorizontalFrame(frame, LAYOUT_FILL_X|PACK_UNIFORM_WIDTH, 0,0,0,0, 0,0,0,0, 0,0);
	switcher=new FXSwitcher(frame, LAYOUT_FILL|FRAME_NONE, 0,0,0,0, 0,0,0,0);*/

	new FXButton(frame, tr("&Dirlist"), GET_ICON(icon_viewtreelist), 0,
		0, BUTTON_NORMAL|LAYOUT_FILL_X);
	/*new FXButton(buttons, tr(""), 0, switcher,
		FXSwitcher::ID_OPEN_SECOND, BUTTON_NORMAL|LAYOUT_FILL_X);*/

	_dirList=new FXDirList(frame, this, mainWindow::ID_DIRLIST,
		TREELIST_SHOWS_BOXES|DIRLIST_SHOWFILES|TREELIST_SINGLESELECT|LAYOUT_FILL);
	_dirList->dropDisable();
	_dirList->setPattern(TFACTORY->plExt());
#if FOX_MINOR>6
	_dirList->setMatchMode(NoEscape|PathName|CaseFold);
#else
	_dirList->setMatchMode(FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD);
#endif

	//_bookmarkList=new bookmark(switcher, this, mainWindow::ID_BOOKMARKS);

	for (int i=0; i<IFACTORY->num(); i++)
		_fileTypes->appendItem(IFACTORY->name(i));

	_fileTypes->setNumVisible((_fileTypes->getNumItems()>5)?5:_fileTypes->getNumItems());

	_width=getApp()->reg().readIntEntry("General Settings", "width");
	if (!_width)
		_width=getWidth();
	else
		setWidth(_width);

	_height=getApp()->reg().readIntEntry("General Settings", "height");
	if (!_height)
		_height=getHeight();
	else
		setHeight(_height);
}

mainWindow::~mainWindow() {

	delete ledRedIcon;
	delete ledYellowIcon;
	delete ledGreenIcon;
}

void mainWindow::create() {

	FXMainWindow::create();

	ledGreenIcon->create();
	ledYellowIcon->create();

	// Load settings
	FXString tmp=getApp()->reg().readStringEntry("General Settings", "sort", "ascending");
	_playList->sorting(tmp.text());

	setOpts();

	const FXbool playMode=getApp()->reg().readBoolEntry("General Settings", "play mode", TRUE);
	 _playMode->setState(playMode);

	setBackend(0, 0);

	// Look at command line arguments
	if (getApp()->getArgc()>1) {
#if defined (_WIN32) && defined (UNICODE)
		int narg;
		FXnchar** arg=CommandLineToArgvW(GetCommandLineW(), &narg);
		const FXchar* argv=convertNc(arg[1]);
		const FXString message=argv;
		delete [] argv;
#else
		const FXString message=(getApp()->getArgv()[1]);
#endif
		setFile(FXPath::absolute(message), true);
	} else
		setFile(getApp()->reg().readStringEntry("General Settings", "file"), false);

	show(PLACEMENT_SCREEN);
}

long mainWindow::onHome(FXObject*, FXSelector, void*) {

	if (_input) {
		const FXString musicDir=_input->getMusicDir();
		FXTRACE((10, "music dir: %s\n", musicDir.text()));
		_dirList->setCurrentFile((musicDir.empty())?getMusicDir():musicDir, TRUE);
	}
	return 1;
}

void mainWindow::setFile(const FXString& file, const bool play) {

	FXTRACE((10, "setFile %s\n", file.text()));

	// Check if requested file/directory exists
	if (!FXStat::exists(file)) {
		handle(this, FXSEL(SEL_COMMAND, ID_HOME), 0);
		return;
	}

	FXint curItem=_playList->getCurrentItem();
	FXint val=_playList->findItem(file);
	if (_playing && (val==curItem))
		return;

	const FXString fName=FXPath::name(file);

	const bool dirSelected=!compare(FXPath::directory(file), _dirList->getCurrentFile());

	if (FXStat::isDirectory(file) || TFACTORY->plExt().find(FXPath::extension(file).lower())>=0) {
		_dirList->setCurrentFile(file, TRUE);
		_playMode->setState(TRUE);
		goto done;
	}

	if (fName.empty())
		goto ok;

	if (_input &&
#if FOX_MINOR>6
	FXPath::match(fName, _input->ext(), NoEscape|PathName|CaseFold))
#else
	FXPath::match(_input->ext(), fName, FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD))
#endif
		goto ok;

	for (int i=0; i<IFACTORY->num(); i++) {
		if (IFACTORY->supports(i, fName.text())) {

			handle(this, FXSEL(SEL_COMMAND, ID_STOP), 0);

			setBackend(i, dirSelected?1:0);
			val=_playList->findItem(file);
			curItem=-1;
			goto ok;
		}
	}

	// file is not supported, ignore message
	return;

ok:
	if (dirSelected) {
		if (val>=0) {
			if (val==curItem)
				handle(this, FXSEL(SEL_COMMAND, ID_PLAYPAUSE), 0);
			else
				_playList->setCurrentItem(val, TRUE);
		}
	} else {
		FXString* fileName=new FXString(file);
		_dirList->setUserData((void*)fileName);
		_dirList->setCurrentFile(FXPath::directory(file), TRUE);
	}

done:
	if (play) {
		if (_loadThread)
			_loadThread->join();
		handle(this, FXSEL(SEL_COMMAND, ID_PLAYPAUSE), 0);
	}
}

FXbool mainWindow::close(FXbool notify) {

	handle(this, FXSEL(SEL_COMMAND, ID_STOP), 0);

	// Save Settings
	getApp()->reg().writeIntEntry("General Settings", "width", _mainView->shown()?getWidth():_width);
	getApp()->reg().writeIntEntry("General Settings", "height", _mainView->shown()?getHeight():_height);

	const char* file=_input->songLoaded();
	getApp()->reg().writeStringEntry("General Settings", "file", file?file:"");
	getApp()->reg().writeBoolEntry("General Settings", "play mode", _playMode->getState());

	getApp()->reg().writeStringEntry("General Settings", "sort", _playList->sorting());

	_bookmarkList->save();

	delete _audio;
	delete _input;

	return FXMainWindow::close(notify);
}

void mainWindow::setOpts() {

	_playList->setAltBackColor(SETTINGS->altCol());
	_bookmarkList->setAltBackColor(SETTINGS->altCol());
}

long mainWindow::onMessage(FXObject*, FXSelector, void* ih) {

	FXTRACE((10, "onMessage\n"));

	FXchar buffer[2048];

	// FIXME cleanup and maybe move outside
#ifdef WIN32
	LPOVERLAPPED overlapped=static_cast<singleApp*>(getApp())->getOverlapped();
	const HANDLE _handle=static_cast<singleApp*>(getApp())->getPipe();
	const FXInputHandle inputHandle=(FXInputHandle)_handle;
#else
	const FXInputHandle inputHandle=(FXInputHandle)(FXival)ih;
#endif

	FXPipe pipe;
	pipe.attach(inputHandle, FXIO::Reading|FXIO::NonBlocking);
	const int n=pipe.readBlock(buffer, 2048);
	pipe.detach();

#ifdef WIN32
	if (GetLastError()==ERROR_BROKEN_PIPE) {
		if (DisconnectNamedPipe(_handle)) {
			FXTRACE((10, "Disconnected from pipe\n"));
		}

		HANDLE event=overlapped->hEvent;

		if (ResetEvent(event)) {
			FXTRACE((10, "Event resetted\n"));
		}

		memset(overlapped, 0, sizeof(OVERLAPPED));
		overlapped->hEvent=event;

		ConnectNamedPipe(_handle, overlapped);
		if (GetLastError()==ERROR_IO_PENDING) {
			FXTRACE((10, "Waiting for connection\n"));
		}/* else {
			char message[512];
			FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, GetLastError(), NULL,message, 512,NULL);
			FXTRACE((10, "GetLastError: %s\n", message));
		}*/
	}
#endif

	if (n) {
		const FXString tmp=FXString(buffer, n);

		FXTRACE((10, "message: %s\n", tmp.text()));

		if (!tmp.empty())
			setFile(tmp, true);
	}

	return 1;
}

long mainWindow::onCmdInfo(FXObject*, FXSelector sel, void*) {

	FXDialogBox* dialog=0;
	switch(FXSELID(sel)) {
	case ID_INFO:
		dialog=new infoDialog(this, _input->getMetaData());
		break;
	case ID_ABOUT:
		dialog=new aboutDialog(this, GET_ICON(icon_logo32));
		break;
	}

	dialog->execute(PLACEMENT_OWNER);
	delete dialog;
	return 1;
}

long mainWindow::onCmdConfig(FXObject*, FXSelector sel, void*) {

	settingsWindow cWin(this, _input);
	if (cWin.execute(PLACEMENT_OWNER)) {
		handle(this, FXSEL(SEL_COMMAND, ID_STOP), 0);
		SETTINGS->save();
		setOpts();
		_audio->setOpts();
		_input->saveSettings();
		if (_playList->getNumItems()) {
			// if settings changes we must reload the song
			_input->close();
			const FXint curItem=_playList->getCurrentItem();
			_playList->killSelection();
			_playList->selectItem(curItem, TRUE);
		}
	} else {
		SETTINGS->load();
		_input->loadSettings();
	}

	return 1;
}

long mainWindow::onCmdDirList(FXObject*, FXSelector, void* autoBk) {

	if (_editMode->getState())
		return 0;

	const FXString curItem=_dirList->getCurrentFile();
	if (curItem.empty())
		return 0;
/*
	if (FXPath::match(_input->ext(), curItem, FILEMATCH_FILE_NAME|FILEMATCH_NOESCAPE|FILEMATCH_CASEFOLD))
		return 0;
*/
	FXTRACE((10, "Set dir %s\n", curItem.text()));

	int items=_playList->load(curItem);

	if (!items)
		return 1;

	items=_input?_playList->filter(_input->ext()):0;
	if (!items) {
		if (_playing || !SETTINGS->autoBk() || !autoBk)
			return 1;
		int bk=0;
		input* i=0;
		do {
			if (bk!=_fileTypes->getCurrentItem()) {
				i=backend(bk);
				items=_playList->filter(i->ext());
				delete i;
			}
		} while (++bk<_fileTypes->getNumItems() && !items);
		if (items)
			setBackend(bk-1, 0);
		else
			return 1;
	}

	_playList->sortItems();
	_playList->killSelection();

	if (!_playing) {
		FXString* fileName=(FXString*)_dirList->getUserData();
		if (fileName) {
			const FXint val=_playList->findItem(*fileName);
			delete fileName;

			_playList->setCurrentItem((val<0)?0:val, TRUE);
			_dirList->setUserData((void*)(FXuval)0);
		} else
			_playList->setCurrentItem(0, TRUE);
	} else if (!compare(_playDir, curItem)) {
		FXint i=_playList->findItem(_input->songLoaded());
		_playList->setCurrentItem(i);
		_playList->makeItemVisible(i);
	}

	return 1;
}

long mainWindow::onRgtClkDirList(FXObject*, FXSelector, void* ptr) {

	FXEvent* event=(FXEvent*)ptr;
	if (event->moved)
		return 1;

	FXTreeItem* item=_dirList->getItemAt(event->win_x, event->win_y);
	if (!item)
		return 0;

	FXMenuPane pane(this);
	new FXLabel(&pane, shrink(item->getText()));
	new FXMenuSeparator(&pane);

	if ((_editMode->getState()) && (_dirList->isItemFile(item)))
		new FXMenuCommand(&pane, tr("&Add to playlist\t\tAdd song to the playlist."),
			GET_ICON(icon_listadd), this, mainWindow::ID_PL_ADD);
	else
		new FXMenuCommand(&pane, tr("&Add to bookmarks\t\tAdd directory to the bookmarks."),
			GET_ICON(icon_bookmark), this, mainWindow::ID_BM_ADD);

	new FXMenuCommand(&pane, tr("Cancel\t\tCancel action"), GET_ICON(icon_actioncancel));

	pane.create();
	setUserData((void*)item);
	pane.popup(0, event->root_x, event->root_y);
	getApp()->runModalWhileShown(&pane);
	return 1;
}

void mainWindow::setPlayButton() {

	const unsigned char* icon;
	const char* label;

	if (_audio->state()!=PLAY) {
		icon=icon_mediaplaybackstart;
		label=tr("Play");
	} else{
		icon=icon_mediaplaybackpause;
		label=tr("Pause");
	}

	_playpause->setIcon(GET_ICON(icon));
	_playpause->setTipText(label);
	_playpause->setHelpText(label);
}

long mainWindow::onCmdPlayPauseSong(FXObject*, FXSelector, void*) {

	if (_audio->state()!=STOP)
		_audio->pause();
	else if (_audio->play(_input)) {
		_playing=true;
		_playDir=_dirList->getCurrentFile();
		notify();
	}
	
	setPlayButton();
	return 1;
}

long mainWindow::onCmdStopSong(FXObject*, FXSelector, void*) {

	if (_audio->stop()) {
		_playing=false;
		_playDir=FXString::null;
		_timeDisplay->setTime(0);
		notify();
		setPlayButton();
	}
	return 1;
}

long mainWindow::onCmdChangeSong(FXObject*, FXSelector sel, void*) {

	if (_playing && compare(_playDir, _dirList->getCurrentFile()))
		return 0;

	FXint idx=_playList->findItem(_input->songLoaded());

	switch (FXSELID(sel)) {
	case ID_PREV:
		idx--;
		FXTRACE((10, "Previous\n"));
		break;
	case ID_NEXT:
		idx++;
		FXTRACE((10, "Next\n"));
		break;
	}

	if ((idx>=0) && (idx<_playList->getNumItems())) {
		if (!_preloaded.empty()) {
			_preload->close();
			_preloaded=FXString::null;
		}
		_playList->setCurrentItem(idx, TRUE);
	}

	return 1;
}

long mainWindow::onCmdSongSelected(FXObject*, FXSelector, void* ptr) {

	const FXint curItem=(FXint)(FXival)ptr;
	const char* songLoaded=_input->songLoaded();
	if (songLoaded && !compare(songLoaded, _playList->getItemText(curItem)))
		return 1;

	_playList->makeItemVisible(curItem);
	const FXString song=_playList->getItemText(curItem);

	if (!_preloaded.empty()) {
		if (compare(_preloaded, song)) {
			_preload->close();
			_preloaded=FXString::null;
		} else {
			input *temp=_input;
			_input=_preload;
			_preload=temp;
			_preload->close();
			_preloaded=FXString::null;
			setDisplay();
			notify();
			return 1;
		}
	}

	if (_loadThread) {
		_loadThread->join();
		delete _loadThread;
	} else {
		getApp()->beginWaitCursor();
		_audio->stop();
		clearDisplay(true);
	}

	load(song, _input);

	return 1;
}

long mainWindow::onCmdSongLoaded(FXObject*, FXSelector, void* ptr) {

	_loadThread->join();

	if (!_preloaded.empty()) {
		delPtr(_loadThread);
		if (ptr && _audio->gapless(_preload)) {
			if (SETTINGS->subtunes())
				_preload->subtune(1);
		} else {
			_preloaded=FXString::null;
		}
		return 1;
	}

	if ((FXint)(FXival)ptr) {
		setDisplay();

		if (SETTINGS->subtunes())
			_input->subtune(1);

		if (_playing) {
			notify();
			handle(this, FXSEL(SEL_COMMAND, ID_PLAYPAUSE), 0);
		}
		FXTRACE((10, "Song loaded\n"));
	} else {
		_playing=false;
		_songInfo->setInfo(0);
		fxmessage("Error loading song\n");
	}

	getApp()->endWaitCursor();
	delPtr(_loadThread);

	return 1;
}

void mainWindow::setDisplay() {

	_timeDisplay->setTime(0);
	_songTime->setTime(_input->time());

	const metaData* data=_input->getMetaData();
	FXString songTitle=data->getInfo(metaData::TITLE);
	if (songTitle.empty())
		songTitle=FXPath::title(_input->songLoaded());

	setTitle(songTitle);

	_songInfo->setInfo(data);
}

void mainWindow::clearDisplay(bool const loading) {

	_songInfo->setText(loading?tr("Loading..."):FXString::null);
	_songInfo->setTipText(FXString::null);
	setTitle(FXString(PACKAGE_STRING));
	_songTime->reset();
	_timeDisplay->reset();
}

long mainWindow::onCmdAddItem(FXObject*, FXSelector sel, void*) {

	const FXTreeItem* item=(FXTreeItem*)getUserData();
	const FXString itemName=_dirList->getItemPathname(item);
	switch(FXSELID(sel)) {
	case ID_BM_ADD:
		_bookmarkList->add(itemName);
		break;
	case ID_PL_ADD:
		_playList->add(itemName);
		break;
	}
	return 1;
}

long mainWindow::onCmdBmGet(FXObject*, FXSelector, void* data) {

	_dirList->setCurrentFile(*(FXString*)data, TRUE);
	_dirList->expandTree(_dirList->getCurrentItem());
	return 1;
}

void mainWindow::load(const FXString& filename, input* const i, void* data) {

	FXTRACE((10, "Loading %s\n", filename.text()));

	FXGUISignal* songLoaded=new FXGUISignal(getApp(), this, mainWindow::ID_SONG_LOADED);
	_loadThread=new loadThread(i, filename, songLoaded);

	_loadThread->start();
}

long mainWindow::onAudioMsg(FXObject*, FXSelector, void* data) {

	switch ((int)data) {
	case PRELOAD_SONG:
		if (_playMode->getState() && (!compare(_playList->sorting(), "ascending")) &&
		_input->gapless() && !compare(_playDir, _dirList->getCurrentFile())) {
			const FXint nextSong=_playList->getCurrentItem()+1;
			if (nextSong<_playList->getNumItems()) {
				_preloaded=_playList->getItemText(nextSong);
				load(_preloaded, _preload);
			}
		}
	case UPD_TIME:
		_timeDisplay->setTime(_audio->seconds());
		return 1;
	case SONG_ENDED:
		if (SETTINGS->subtunes() && (_input->subtune()<_input->subtunes())) {
			handle(this, FXSEL(SEL_COMMAND, ID_NEXT_SUBTUNE), 0);
			return 1;
		}

		if (_playMode->getState()) {
			FXint item=0;
			if (!compare(_playDir, _dirList->getCurrentFile()))
				item=_playList->findItem(_input->songLoaded())+1;

			if (item<_playList->getNumItems()) {
				_playList->setCurrentItem(item, TRUE);
				return 1;
			}
		}
	case OUTPUT_ERROR:
		handle(this, FXSEL(SEL_COMMAND, ID_STOP), 0);

		return 1;
	}

	return 0;
}

long mainWindow::onCmdUIMode(FXObject* obj, FXSelector, void* ptr) {

	_mainView->handle(obj, FXSEL(SEL_COMMAND, ID_TOGGLESHOWN), ptr);
	if (_mainView->shown()) {
		setDecorations(DECOR_ALL);
		resize(_width, _height);
		_statusbar->setCornerStyle(TRUE);
	} else {
		_width=getWidth();
		_height=getHeight();
		setDecorations(DECOR_TITLE|DECOR_BORDER|DECOR_MINIMIZE|DECOR_CLOSE);
		resize(getDefaultWidth(), getDefaultHeight());
		_statusbar->setCornerStyle(FALSE);
	}

	return 1;
}

long mainWindow::onCmdFiletype(FXObject*, FXSelector, void* ptr) {

	setBackend((FXint)(FXival)ptr, 2);
	return 1;
}

input* mainWindow::backend(const int val) {
	return IFACTORY->get(val);
}

void mainWindow::setBackend(const int val, const int refresh) {

	FXTRACE((10, "setBackend %d\n", val));
	if (val<0 || val>_fileTypes->getNumItems())
		return;

	_fileTypes->setCurrentItem(val);

	if (_input) {
		_bookmarkList->save();
		clearDisplay(false);
		delete _input;
		delete _preload;
	}

	_input=backend(val);
	_preload=(_input->gapless())?backend(val):0;

	_bookmarkList->load(IFACTORY->name(val));

	if (refresh)
		handle(this, FXSEL(SEL_CHANGED, ID_DIRLIST), (void*)(FXival)((refresh==2)?0:1));
}

long mainWindow::onUpdFiletype(FXObject* sender, FXSelector, void*) {

	return sender->handle(this,
		((_audio->state()==STOP) && (!_loadThread) && (!_editMode->getState()))?
			FXSEL(SEL_COMMAND, ID_ENABLE):FXSEL(SEL_COMMAND, ID_DISABLE),
		0);
}

long mainWindow::onUpdLed(FXObject* sender, FXSelector, void*) {

	FXLabel* led=(FXLabel*)sender;
	switch (_audio->state()) {
	case STOP:
		led->setIcon(ledRedIcon);
		led->setHelpText(tr("Stopped"));
		break;
	case PLAY:
		led->setIcon(ledGreenIcon);
		led->setHelpText(tr("Playing"));
		break;
	case PAUSE:
		led->setIcon(ledYellowIcon);
		led->setHelpText(tr("Paused"));
		break;
	}
	return 1;
}

long mainWindow::onUpdSubtune(FXObject* sender, FXSelector sel, void*) {

	int tune, tunes;
	switch (FXSELID(sel)) {
	case ID_PLSONG:
		if (_playing && compare(_playDir, _dirList->getCurrentFile()))
			return 0;
		tunes=_playList->getNumItems();
		tune=1+_playList->getCurrentItem();
		break;
	case ID_SUBTUNE:
		tune=_input->subtune();
		tunes=_input->subtunes();
		break;
	}

	FXLabel* label=(FXLabel*)sender;
	label->setText(FXStringFormat("%.2d/%.2d", tune, tunes));
	return 1;
}

long mainWindow::onCmdSubtune(FXObject*, FXSelector sel, void*) {

	if (_input->subtunes()<=1)
		return 0;

	FXuint i=_input->subtune();
	switch (FXSELID(sel)) {
	case ID_PREV_SUBTUNE:
		i--;
		break;
	case ID_NEXT_SUBTUNE:
		i++;
		break;
	}

	if ((i<1) || (i>_input->subtunes()))
		return 0;

	_audio->stop();

	if (_input->subtune(i))
		setDisplay();

	if (_playing)
		handle(this, FXSEL(SEL_COMMAND, ID_PLAYPAUSE), 0);
	return 1;
}

long mainWindow::onCmdPlBack(FXObject*, FXSelector, void* ptr) {

	_dirList->setCurrentFile(FXPath::directory(_input->songLoaded()), TRUE);
	return 1;
}

long mainWindow::onCmdPlEdit(FXObject*, FXSelector, void* ptr) {

	if (ptr) {
		_playList->clearItems();
		_playList->dropEnable();
		_dirList->setPattern(_input->ext());
	} else {
		_playList->dropDisable();
		_dirList->setPattern(TFACTORY->plExt());
		handle(this, FXSEL(SEL_CHANGED, ID_DIRLIST), 0);
	}

	return 1;
}

long mainWindow::onCmdPlSave(FXObject* sender, FXSelector, void*) {

	const FXint n=_playList->getNumItems();
	if (!n)
		return 1;

	FXString filename=FXFileDialog::getSaveFilename(this, tr("Save playlist"), "", TFACTORY->plExt());
	if (filename.empty())
		return 1;

	if (FXStat::exists(filename))
		if (FXMessageBox::question(this, MBOX_YES_NO, tr("Warning"), tr("File exists, owerwrite?"))==MBOX_CLICKED_NO)
			return 1;

	if (!_playList->save(filename))
		FXMessageBox::error(this, MBOX_OK, tr("Error"), tr("Error saving playlist"));

	return 1;
}

long mainWindow::onCmdVol(FXObject*, FXSelector, void* data) {

	_audio->volume((FXint)(FXival)data);
	return 1;
}

void mainWindow::notify() {
#ifdef _WIN32
	if (SETTINGS->nowPlaying()) {
		FXTRACE((10, "notify %s\n", _playing?"play":"stop"));
		const metaData* data=_input->getMetaData();
		nowPlaying(_playing, SETTINGS->wlm(), data->getInfo(metaData::TITLE), data->getInfo(metaData::ARTIST), data->getInfo(metaData::ALBUM));
	}
#endif
}
